@@ -124,7 +124,7 @@
  *
  * `AdwAlertDialog` supports adding responses in UI definitions by via the
  * `<responses>` element that may contain multiple `<response>` elements, each
- * respresenting a response.
+ * representing a response.
  *
  * Each of the `<response>` elements must have the `id` attribute specifying the
  * response ID. The contents of the element are used as the response label.
@@ -351,18 +351,30 @@ static void
 adw_alert_dialog_map (GtkWidget *widget)
 {
   AdwAlertDialog *self = ADW_ALERT_DIALOG (widget);
-  AdwAlertDialogPrivate *priv = adw_alert_dialog_get_instance_private (self);
-  GtkWidget *focus;
   GtkWidget *window;
 
   GTK_WIDGET_CLASS (adw_alert_dialog_parent_class)->map (widget);
 
   window = adw_dialog_get_window (ADW_DIALOG (self));
 
-  if (window)
+  if (window) {
     gtk_widget_add_css_class (window, "alert");
+    gtk_widget_grab_focus (widget);
+  }
+}
+
+static gboolean
+adw_alert_dialog_grab_focus (GtkWidget *widget)
+{
+  AdwAlertDialog *self = ADW_ALERT_DIALOG (widget);
+  AdwAlertDialogPrivate *priv = adw_alert_dialog_get_instance_private (self);
+  GtkWidget *focus;
+  gboolean ret;
 
   focus = adw_dialog_get_focus (ADW_DIALOG (self));
+
+  ret = GTK_WIDGET_CLASS (adw_alert_dialog_parent_class)->grab_focus (widget);
+
   if (!focus) {
     GtkWidget *default_widget;
     GList *l;
@@ -373,14 +385,12 @@ adw_alert_dialog_map (GtkWidget *widget)
       if (GTK_IS_LABEL (focus) && !gtk_label_get_current_uri (GTK_LABEL (focus)))
         gtk_label_select_region (GTK_LABEL (focus), 0, 0);
 
-      return;
+      return TRUE;
     }
 
     default_widget = adw_dialog_get_default_widget (ADW_DIALOG (self));
-    if (default_widget) {
-      gtk_widget_grab_focus (default_widget);
-      return;
-    }
+    if (default_widget)
+      return gtk_widget_grab_focus (default_widget);
 
     for (l = g_list_last (priv->responses); l; l = l->prev) {
       ResponseInfo *response = l->data;
@@ -388,10 +398,11 @@ adw_alert_dialog_map (GtkWidget *widget)
       if (!response->enabled)
         continue;
 
-      gtk_widget_grab_focus (response->button);
-      return;
+      return gtk_widget_grab_focus (response->button);
     }
   }
+
+  return ret;
 }
 
 static GtkSizeRequestMode
@@ -409,9 +420,9 @@ measure_heading (GtkWidget      *widget,
                  int            *minimum_baseline,
                  int            *natural_baseline)
 {
-  AdwAlertDialog *self = ADW_ALERT_DIALOG (gtk_widget_get_ancestor (widget, ADW_TYPE_ALERT_DIALOG));
+  AdwAlertDialog *self = g_object_get_data (G_OBJECT (widget), "-adw-alert-dialog");
   AdwAlertDialogPrivate *priv = adw_alert_dialog_get_instance_private (self);
-  int large_min, large_nat, small_min, small_nat;
+  int large_min, large_nat, small_min, small_nat, for_size_large_label;
 
   if (priv->is_short) {
     gtk_widget_measure (priv->heading_label_small, orientation, for_size,
@@ -425,7 +436,14 @@ measure_heading (GtkWidget      *widget,
     return;
   }
 
-  gtk_widget_measure (priv->heading_label, orientation, for_size,
+  if (orientation == GTK_ORIENTATION_HORIZONTAL) {
+    gtk_widget_measure (priv->heading_label, GTK_ORIENTATION_HORIZONTAL, -1,
+                        &for_size_large_label, NULL, NULL, NULL);
+  } else {
+    for_size_large_label = -1;
+  }
+
+  gtk_widget_measure (priv->heading_label, orientation, for_size_large_label,
                       &large_min, &large_nat, NULL, NULL);
   gtk_widget_measure (priv->heading_label_small, orientation, for_size,
                       &small_min, &small_nat, NULL, NULL);
@@ -466,7 +484,7 @@ allocate_heading (GtkWidget *widget,
                   int        height,
                   int        baseline)
 {
-  AdwAlertDialog *self = ADW_ALERT_DIALOG (gtk_widget_get_ancestor (widget, ADW_TYPE_ALERT_DIALOG));
+  AdwAlertDialog *self = g_object_get_data (G_OBJECT (widget), "-adw-alert-dialog");
   AdwAlertDialogPrivate *priv = adw_alert_dialog_get_instance_private (self);
   gboolean small_label;
 
@@ -558,7 +576,7 @@ measure_responses (GtkWidget      *widget,
                    int            *minimum_baseline,
                    int            *natural_baseline)
 {
-  AdwAlertDialog *self = ADW_ALERT_DIALOG (gtk_widget_get_ancestor (widget, ADW_TYPE_ALERT_DIALOG));
+  AdwAlertDialog *self = g_object_get_data (G_OBJECT (widget), "-adw-alert-dialog");
   AdwAlertDialogPrivate *priv = adw_alert_dialog_get_instance_private (self);
 
   if (orientation == GTK_ORIENTATION_HORIZONTAL) {
@@ -657,7 +675,7 @@ measure_child (GtkWidget      *widget,
                int            *min_baseline,
                int            *nat_baseline)
 {
-  AdwAlertDialog *self = ADW_ALERT_DIALOG (gtk_widget_get_ancestor (widget, ADW_TYPE_ALERT_DIALOG));
+  AdwAlertDialog *self = g_object_get_data (G_OBJECT (widget), "-adw-alert-dialog");
   AdwAlertDialogPrivate *priv = adw_alert_dialog_get_instance_private (self);
   int max_size, min_size, base_nat;
 
@@ -699,16 +717,7 @@ measure_child (GtkWidget      *widget,
                                               DIALOG_MAX_WIDTH,
                                               gtk_widget_get_settings (widget));
 
-    if (priv->is_short) {
-      min_size = MAX (min_size, regular_max_size);
-      max_size = adw_length_unit_to_px (ADW_LENGTH_UNIT_SP,
-                                        DIALOG_MAX_WIDE_WIDTH,
-                                        gtk_widget_get_settings (widget));
-
-      max_size = MIN (max_size, wide_nat);
-    } else if (priv->prefer_wide_layout) {
-      min_size = MAX (min_size, regular_max_size);
-
+    if (priv->prefer_wide_layout || priv->is_short) {
       max_size = adw_length_unit_to_px (ADW_LENGTH_UNIT_SP,
                                         DIALOG_MAX_WIDE_WIDTH,
                                         gtk_widget_get_settings (widget));
@@ -741,7 +750,7 @@ allocate_child (GtkWidget *widget,
                 int        height,
                 int        baseline)
 {
-  AdwAlertDialog *self = ADW_ALERT_DIALOG (gtk_widget_get_ancestor (widget, ADW_TYPE_ALERT_DIALOG));
+  AdwAlertDialog *self = g_object_get_data (G_OBJECT (widget), "-adw-alert-dialog");
   AdwAlertDialogPrivate *priv = adw_alert_dialog_get_instance_private (self);
 
   gtk_widget_allocate (priv->window_handle, width, height, baseline, NULL);
@@ -880,6 +889,7 @@ adw_alert_dialog_class_init (AdwAlertDialogClass *klass)
   object_class->finalize = adw_alert_dialog_finalize;
 
   widget_class->map = adw_alert_dialog_map;
+  widget_class->grab_focus = adw_alert_dialog_grab_focus;
 
   dialog_class->closed = adw_alert_dialog_closed;
 
@@ -1064,6 +1074,10 @@ adw_alert_dialog_init (AdwAlertDialog *self)
 
   gtk_widget_init_template (GTK_WIDGET (self));
 
+  g_object_set_data (G_OBJECT (priv->contents), "-adw-alert-dialog", self);
+  g_object_set_data (G_OBJECT (priv->heading_bin), "-adw-alert-dialog", self);
+  g_object_set_data (G_OBJECT (priv->response_area), "-adw-alert-dialog", self);
+
   gtk_widget_set_layout_manager (priv->contents,
                                  gtk_custom_layout_new (adw_widget_get_request_mode,
                                                         measure_child,
@@ -1826,14 +1840,14 @@ adw_alert_dialog_set_extra_child (AdwAlertDialog *self,
   g_return_if_fail (ADW_IS_ALERT_DIALOG (self));
   g_return_if_fail (child == NULL || GTK_IS_WIDGET (child));
 
-  if (child)
-    g_return_if_fail (gtk_widget_get_parent (child) == NULL);
-
   priv = adw_alert_dialog_get_instance_private (self);
 
   if (child == priv->child)
     return;
 
+  if (child)
+    g_return_if_fail (gtk_widget_get_parent (child) == NULL);
+
   priv->child = child;
   adw_bin_set_child (ADW_BIN (priv->child_bin), child);
   gtk_widget_set_visible (priv->child_bin, child != NULL);
@@ -2450,7 +2464,7 @@ choose_cancelled_cb (GCancellable *cancellable,
  * @parent: (nullable): the parent widget
  * @cancellable: (nullable): a `GCancellable` to cancel the operation
  * @callback: (scope async): a callback to call when the operation is complete
- * @user_data: (closure callback): data to pass to @callback
+ * @user_data: data to pass to @callback
  *
  * This function shows @self to the user.
  *
@@ -2507,3 +2521,4 @@ adw_alert_dialog_choose_finish (AdwAlertDialog *self,
 
   return g_quark_to_string (id);
 }
+
